Unit ProtMenu;
{$F+}
{ ---------------------------------------------------------------------- }
{ This unit provides basic support for pop-up windows and menus in DOS   }
{ protected mode.                                                        }
{ ---------------------------------------------------------------------- }

Interface

Uses OPCrt;

Type
  Rect = Record
    Active      : Boolean;
    X1,Y1,X2,Y2 : Integer;
    AUL,AUR,ABL,ABR,AHZ,AVT : String[1];
    PUL,PUR,PBL,PBR,PHZ,PVT : String[1];
    PFore,PBack : Byte;
    AFore,ABack : Byte;
  End;

Const
  ScreenWidth = 132;
{ ------------------------ Keys ------------------------ }
  Key_Null    = #0;
  Key_Tab     = #9;
  Key_Enter   = #13;
  Key_Esc     = #27;
  Key_F1      = #59;
  Key_Alt_F5  = #108;
  Key_Alt_F10 = #113;
  Key_Up      = #72;
  Key_Down    = #80;
  Key_Left    = #75;
  Key_Right   = #80;
  Key_PgUp    = #73;
  Key_PgDn    = #81;
  Key_Home    = #71;
  Key_End     = #79;

Type
  BorderType = (SingleBorder,DoubleBorder);
  ScreenStr  = String[ScreenWidth];
  PSelection = ^DSelection;
  DSelection = Record
    X,Y       : Integer;
    Width     : Integer;
    SelStr    : ScreenStr;
    HotChar   : Byte;
    Protected : Boolean;
    Index     : Integer;
    Next      : PSelection;
    Last      : PSelection;
  End;
  DWindow = Object
    X1,Y1,X2,Y2 : Integer;
    Size        : Word;
    WindowAttr  : Byte;
    BorderAttr  : Byte;
    CaptionAttr : Byte;
    Caption     : ScreenStr;
    Border      : BorderType;
    WindowPtr   : Pointer;
    Visible     : Boolean;
    Shadow      : Boolean;
    Constructor Init(_X1,_Y1,_X2,_Y2: Integer;
                     WFore,WBack,BFore,BBack,CFore,CBack: Byte;
                     BType: BorderType; Cap: ScreenStr; Shad: Boolean);
    Procedure   Display; Virtual;
    Procedure   Remove; Virtual;
    Destructor  Done;
  End;
  DMenu = Object(DWindow)
    SelHead        : PSelection;
    SelTail        : PSelection;
    NumSelections  : Integer;
    CurSel         : Integer;
    OldSel         : Integer;
    HotCharAttr    : Byte;
    SelAttr        : Byte;
    SelHotCharAttr : Byte;
    ProtectedAttr  : Byte;
    SelProtAttr    : Byte;
    Constructor Init(_X1,_Y1,_X2,_Y2: Integer;
                     WFore,WBack,BFore,BBack,CFore,CBack,HCFore,HCBack,
                     SFore,SBack,SHCFore,SHCBack,PFore,PBack,SPFore,SPBack: Byte;
                     BType: BorderType; Cap: ScreenStr; Shad: Boolean);
    Procedure   Display; Virtual;
    Procedure   AddSelection(_X,_Y,SelWidth: Integer; HotCh: Byte;
                             Protect: Boolean; St: ScreenStr);
    Procedure   SetSelection(Sel: Integer);
    Procedure   Execute;
    Procedure   Remove; Virtual;
    Destructor  Done; Virtual;
    Function    GetSelection(Sel: Integer): PSelection;
  End;

Const
  WindowRect: Rect = (Active : False;
                      X1     : 1;
                      Y1     : 1;
                      X2     : 1;
                      Y2     : 1;
                      AUL    : '';
                      AUR    : '';
                      ABL    : '';
                      ABR    : '';
                      AHZ    : '';
                      AVT    : '';
                      PUL    : '';
                      PUR    : '';
                      PBL    : '';
                      PBR    : '';
                      PHZ    : '';
                      PVT    : '';
                      PFore  : LightGray;
                      PBack  : Black;
                      AFore  : LightGray;
                      ABack  : Black);

Var
  MaxX   : Integer;
  MaxY   : Integer;

Procedure WriteAt(X,Y: Integer; Fore,Back: Byte; _St: String);
Procedure WriteAtAttr(X,Y: Integer; Attr: Byte; _St: String);
Procedure ShadeAt(X,Y: Integer; Len: Byte);
Procedure WriteFrame(Var R: Rect);
Procedure ProtMenuInit;

Implementation

Procedure WriteAt(X,Y: Integer; Fore,Back: Byte; _St: String);
Var Addr: Word;
Begin
  If _St = '' Then Exit;
  Inc(X,Lo(WindMin));
  Inc(Y,Hi(WindMin));
  Addr := ((Y - 1) * MaxX + X - 1) * 2;
  Asm
    CLI
    CLD
    MOV    AX,0B800h
    MOV    ES,AX
    PUSH   DS
    MOV    AX,SS
    MOV    DS,AX
    LEA    SI,BYTE PTR _St
    MOV    DI,WORD PTR Addr
    SUB    CX,CX
    LODSB
    MOV    CL,AL
    MOV    BL,BYTE PTR Fore
    MOV    BH,BYTE PTR Back
    SHL    BH,4
    AND    BL,0Fh
    OR     BL,BH
@L1:
    MOVSB
    MOV    AL,BL
    STOSB
    LOOP   @L1
    POP    DS
    STI
  End; { Asm }
End; { WriteAt }


Procedure WriteAtAttr(X,Y: Integer; Attr: Byte; _St: String);
Var Addr: Word;
Begin
  If _St = '' Then Exit;
  Inc(X,Lo(WindMin));
  Inc(Y,Hi(WindMin));
  Addr := ((Y - 1) * MaxX + X - 1) * 2;
  Asm
    CLI
    CLD
    MOV    AX,0B800h
    MOV    ES,AX
    PUSH   DS
    MOV    AX,SS
    MOV    DS,AX
    LEA    SI,BYTE PTR _St
    MOV    DI,WORD PTR Addr
    SUB    CX,CX
    LODSB
    MOV    CL,AL
    MOV    BL,BYTE PTR Attr
@L1:
    MOVSB
    MOV    AL,BL
    STOSB
    LOOP   @L1
    POP    DS
    STI
  End; { Asm }
End; { WriteAtAttr }


Procedure ShadeAt(X,Y: Integer; Len: Byte);
Var Addr: Word;
Begin
  Inc(X,Lo(WindMin));
  Inc(Y,Hi(WindMin));
  Addr := ((Y - 1) * MaxX + X - 1) * 2;
  Asm
    CLI
    CLD
    MOV    DI,WORD PTR Addr
    MOV    AX,0B800h
    MOV    ES,AX
    SUB    CX,CX
    MOV    CL,BYTE PTR Len
    MOV    AL,8
@L1:
    INC    DI
    STOSB
    LOOP   @L1
    STI
  End; { Asm }
End; { ShadeAt }


Procedure WriteFrame(Var R: Rect);
Var I: Integer;
Begin
  If R.Active Then
  Begin
    WriteAt(R.X1,R.Y1,R.AFore,R.ABack,R.AUL);
    WriteAt(R.X2,R.Y1,R.AFore,R.ABack,R.AUR);
    WriteAt(R.X1,R.Y2,R.AFore,R.ABack,R.ABL);
    WriteAt(R.X2,R.Y2,R.AFore,R.ABack,R.ABR);
    For I := R.X1 + 1 To R.X2 - 1 Do
    Begin
      WriteAt(I,R.Y1,R.AFore,R.ABack,R.AHZ);
      WriteAt(I,R.Y2,R.AFore,R.ABack,R.AHZ);
    End; { For I }
    For I := R.Y1 + 1 To R.Y2 - 1 Do
    Begin
      WriteAt(R.X1,I,R.AFore,R.ABack,R.AVT);
      WriteAt(R.X2,I,R.AFore,R.ABack,R.AVT);
    End; { For I }
  End
  Else
  Begin
    WriteAt(R.X1,R.Y1,R.PFore,R.PBack,R.PUL);
    WriteAt(R.X2,R.Y1,R.PFore,R.PBack,R.PUR);
    WriteAt(R.X1,R.Y2,R.PFore,R.PBack,R.PBL);
    WriteAt(R.X2,R.Y2,R.PFore,R.PBack,R.PBR);
    For I := R.X1 + 1 To R.X2 - 1 Do
    Begin
      WriteAt(I,R.Y1,R.PFore,R.PBack,R.PHZ);
      WriteAt(I,R.Y2,R.PFore,R.PBack,R.PHZ);
    End; { For I }
    For I := R.Y1 + 1 To R.Y2 - 1 Do
    Begin
      WriteAt(R.X1,I,R.PFore,R.PBack,R.PVT);
      WriteAt(R.X2,I,R.PFore,R.PBack,R.PVT);
    End; { For I }
  End;
  TextColor(Black);
  TextBackground(Cyan);
End; { WriteFrame }

{------------------------------- }
{ DWindow                        }
{------------------------------- }

Constructor DWindow.Init(_X1,_Y1,_X2,_Y2: Integer;
                        WFore,WBack,BFore,BBack,CFore,CBack: Byte;
                        BType: BorderType; Cap: ScreenStr; Shad: Boolean);
Var Swap: Integer;
Begin
  If _X2 < _X1 Then
  Begin
    Swap := _X2;
    _X2  := _X1;
    X1   := Swap;
  End;
  If _Y2 < _Y1 Then
  Begin
    Swap := _Y2;
    _Y2  := _Y1;
    Y1   := Swap;
  End;
  X1          := _X1;
  Y1          := _Y1;
  X2          := _X2;
  Y2          := _Y2;
  WindowAttr  := (WFore And $F) + (WBack Shl 4);
  BorderAttr  := (BFore And $F) + (BBack Shl 4);
  CaptionAttr := (CFore And $F) + (CBack Shl 4);
  Caption     := Cap;
  Border      := BType;
  Shadow      := Shad;
  WindowPtr   := Nil;
  Visible     := False;
  If Shadow
   Then Size := (Y2 - Y1 + 2) * (X2 - X1 + 2) * 2
   Else Size := (Y2 - Y1 + 1) * (X2 - X1 + 1) * 2;
End; { DWindow.Init }


Procedure DWindow.Display;
Var I: Integer;
Begin
  If Visible Then Exit;
  GetMem(WindowPtr,Size);
  If Shadow Then
  Begin
    For I := Y1 To Y2 + 1 Do
    Begin
      Move(Mem[$B800:((I - 1) * MaxX + X1 - 1) * 2],
           Mem[Seg(WindowPtr^):Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 2) * 2],
           (X2 - X1 + 2) * 2);
    End; { For I }
  End
  Else
  Begin
    For I := Y1 To Y2 Do
    Begin
      Move(Mem[$B800:((I - 1) * MaxX + X1 - 1) * 2],
           Mem[Seg(WindowPtr^):Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 1) * 2],
           (X2 - X1 + 1) * 2);
    End; { For I }
  End;
  If Shadow Then
  Begin
    For I := Y1 + 1 To Y2 Do ShadeAt(X2 + 1,I,1);
    ShadeAt(X1 + 1,Y2 + 1,X2 - X1 + 1);
  End;
  TextAttr := BorderAttr;
  WindowRect.X1 := X1;
  WindowRect.Y1 := Y1;
  WindowRect.X2 := X2;
  WindowRect.Y2 := Y2;
  WindowRect.AFore := BorderAttr And $F;
  WindowRect.ABack := BorderAttr Shr 4;
  WindowRect.PFore := BorderAttr And $F;
  WindowRect.PBack := BorderAttr Shr 4;
  Case Border Of
    SingleBorder: WindowRect.Active := False;
    DoubleBorder: WindowRect.Active := True;
  End; { Case }
  WriteFrame(WindowRect);
  Window(X1,Y1,X2,Y2);
  If Caption <> '' Then
  Begin
    TextAttr := CaptionAttr;
    WriteAtAttr(((X2 - X1 + 1) Div 2) - ((Length(Caption) Div 2)),1,
     CaptionAttr,Caption);
  End;
  TextAttr := WindowAttr;
  Window(X1 + 1,Y1 + 1,X2 - 1,Y2 - 1);
  ClrScr;
  Visible := True;
End; { DWindow.Display }


Procedure DWindow.Remove;
Var I: Integer;
Begin
  If (WindowPtr = Nil) Or Not Visible Then Exit;
  Window(1,1,MaxX,MaxY);
  If Shadow Then
  Begin
    For I := Y1 To Y2 + 1 Do
    Begin
      Move(Mem[Seg(WindowPtr^):Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 2) * 2],
           Mem[$B800:((I - 1) * MaxX + X1 - 1) * 2],
           (X2 - X1 + 2) * 2);
    End; { For I }
  End
  Else
  Begin
    For I := Y1 To Y2 Do
    Begin
      Move(Mem[Seg(WindowPtr^):Ofs(WindowPtr^) + (I - Y1) * (X2 - X1 + 1) * 2],
           Mem[$B800:((I - 1) * MaxX + X1 - 1) * 2],
           (X2 - X1 + 1) * 2);
    End; { For I }
  End;
  FreeMem(WindowPtr,Size);
  Visible := False;
  WindowPtr := Nil;
End; { DWindow.Remove }


Destructor DWindow.Done;
Begin
  If Visible Then Remove;
End; { DWindow.Done }

{------------------------------- }
{ DMenu                          }
{------------------------------- }

Constructor DMenu.Init(_X1,_Y1,_X2,_Y2: Integer;
                       WFore,WBack,BFore,BBack,CFore,CBack,HCFore,HCBack,
                       SFore,SBack,SHCFore,SHCBack,PFore,PBack,SPFore,SPBack: Byte;
                       BType: BorderType; Cap: ScreenStr; Shad: Boolean);
Begin
  Inherited Init(_X1,_Y1,_X2,_Y2,WFore,WBack,BFore,BBack,CFore,CBack,BType,Cap,Shad);
  HotCharAttr    := (HCFore  And $F) + (HCBack  Shl 4);
  SelAttr        := (SFore   And $F) + (SBack   Shl 4);
  SelHotCharAttr := (SHCFore And $F) + (SHCBack Shl 4);
  ProtectedAttr  := (PFore   And $F) + (PBack   Shl 4);
  SelProtAttr    := (SPFore  And $F) + (SPBack  Shl 4);
  SelHead        := Nil;
  SelTail        := Nil;
  NumSelections  := 0;
  CurSel         := 0;
  OldSel         := 0;
End; { DMenu.Init }


Procedure DMenu.AddSelection(_X,_Y,SelWidth: Integer; HotCh: Byte;
                             Protect: Boolean; St: ScreenStr);
Var SelWork: PSelection;
Begin
  If SelTail = Nil Then
  Begin
    New(SelHead);
    SelTail := SelHead;
    SelTail^.Next := SelTail;
    SelTail^.Last := SelTail;
  End
  Else
  Begin
    New(SelWork);
    SelTail^.Next := SelWork;
    SelHead^.Last := SelWork;
    SelWork^.Next := SelHead;
    SelWork^.Last := SelTail;
    SelTail       := SelWork;
  End;
  Inc(NumSelections);
  SelTail^.X         := _X;
  SelTail^.Y         := _Y;
  SelTail^.Width     := SelWidth;
  SelTail^.HotChar   := HotCh;
  SelTail^.Protected := Protect;
  SelTail^.SelStr    := St;
  SelTail^.Index     := NumSelections;
  CurSel := 0;
End; { DMenu.AddSelection }


Function DMenu.GetSelection(Sel: Integer): PSelection;
Var SelWork: PSelection;
Begin
  If (SelHead = Nil) Or (Sel < 1) Or (Sel > NumSelections) Then
  Begin
    GetSelection := Nil;
    Exit;
  End;
  SelWork := SelHead;
  While Sel > 1 Do
  Begin
    Dec(Sel);
    SelWork := SelWork^.Next;
  End; { While }
  GetSelection := SelWork;
End; { DMenu.GetSelection }


Procedure DMenu.SetSelection(Sel: Integer);
Var
  SelWork : PSelection;
  St      : ScreenStr;

Begin
  If (Sel <> OldSel) And (SelHead <> Nil) Then
  Begin
    CurSel  := Sel;
    If (OldSel > 0) And (OldSel <= NumSelections) Then
    Begin
      Sel     := OldSel;
      SelWork := SelHead;
      While Sel > 1 Do
      Begin
        Dec(Sel);
        SelWork := SelWork^.Next;
      End; { While }
      St := SelWork^.SelStr;
      While Length(St) < SelWork^.Width Do St := St + ' ';
      If Not SelWork^.Protected Then
      Begin
        WriteAtAttr(SelWork^.X,SelWork^.Y,WindowAttr,St);
        If SelWork^.HotChar > 0 Then
         WriteAtAttr(SelWork^.X + SelWork^.HotChar - 1,SelWork^.Y,HotCharAttr,Copy(St,SelWork^.HotChar,1));
      End
      Else
      Begin
        WriteAtAttr(SelWork^.X,SelWork^.Y,ProtectedAttr,St);
        If SelWork^.HotChar > 0 Then
         WriteAtAttr(SelWork^.X + SelWork^.HotChar - 1,SelWork^.Y,HotCharAttr,Copy(St,SelWork^.HotChar,1));
      End;
    End;
    SelWork := SelHead;
    Sel     := CurSel;
    OldSel  := Sel;
    If (Sel < 1) Or (Sel > NumSelections) Then Exit;
    While Sel > 1 Do
    Begin
      Dec(Sel);
      SelWork := SelWork^.Next;
    End; { While }
    St := SelWork^.SelStr;
    While Length(St) < SelWork^.Width Do St := St + ' ';
    If Not SelWork^.Protected Then
    Begin
      WriteAtAttr(SelWork^.X,SelWork^.Y,SelAttr,St);
      If SelWork^.HotChar > 0 Then
       WriteAtAttr(SelWork^.X + SelWork^.HotChar - 1,SelWork^.Y,SelHotCharAttr,Copy(St,SelWork^.HotChar,1));
    End
    Else
    Begin
      WriteAtAttr(SelWork^.X,SelWork^.Y,SelProtAttr,St);
      If SelWork^.HotChar > 0 Then
       WriteAtAttr(SelWork^.X + SelWork^.HotChar - 1,SelWork^.Y,SelHotCharAttr,Copy(St,SelWork^.HotChar,1));
    End;
  End;
End; { DMenu.SetSelection }


Procedure DMenu.Display;
Var
  SelWork : PSelection;
  Found   : Boolean;
  Index   : Integer;
  St      : ScreenStr;

Begin
  If SelHead <> Nil Then
  Begin
    Inherited Display;
    SelWork := SelHead;
    Repeat
      St := SelWork^.SelStr;
      While Length(St) < SelWork^.Width Do St := St + ' ';
      If SelWork^.Protected Then
       WriteAtAttr(SelWork^.X,SelWork^.Y,ProtectedAttr,St)
      Else
      Begin
        WriteAtAttr(SelWork^.X,SelWork^.Y,WindowAttr,St);
        If SelWork^.HotChar > 0 Then
         WriteAtAttr(SelWork^.X + SelWork^.HotChar - 1,SelWork^.Y,HotCharAttr,Copy(St,SelWork^.HotChar,1));
      End;
      SelWork := SelWork^.Next;
    Until SelWork = SelHead;
    If CurSel = 0 Then CurSel := 1;
    SetSelection(CurSel);
  End;
End; { DMenu.Display }


Destructor DMenu.Done;
Var SelWork,SelWork1: PSelection;
Begin
  If SelHead <> Nil Then
  Begin
    SelWork  := SelHead;
    SelWork1 := SelHead;
    Repeat
      SelWork := SelWork^.Next;
      Dispose(SelWork1);
      SelWork1 := SelWork;
    Until SelWork = SelHead;
    SelHead       := Nil;
    SelTail       := Nil;
    NumSelections := 0;
    CurSel        := 0;
  End;
End; { DMenu.Done }


Procedure DMenu.Execute;
Var
  Finished  : Boolean;
  Key       : Char;
  ScanCode  : Char;
  SelWork   : PSelection;
  SelWork1  : PSelection;
  Sel       : Integer;
  FoundChar : Boolean;

Begin
  If SelHead = Nil Then Exit;
  If Not Visible Then Display;
  Finished := False;
  While Not Finished Do
  Begin
    Key := UpCase(ReadKey);
    Case Key Of
      Key_Esc:
      Begin
        SetSelection(0);
        Finished := True;
      End;
      Key_Enter:
      Begin
        SelWork := SelHead;
        Sel := 1;
        While Sel < CurSel Do
        Begin
          Inc(Sel);
          SelWork := SelWork^.Next;
        End; { While }
        If Not SelWork^.Protected Then Finished := True;
      End;
      Key_Null:
      Begin
        ScanCode := ReadKey;
        Case ScanCode Of
          Key_Up: Begin
{                    Dec(CurSel);
                    If CurSel < 1 Then CurSel := NumSelections;}
                    Sel := CurSel - 1;
                    While (Sel <> CurSel) And GetSelection(Sel)^.Protected Do
                    Begin
                      Dec(Sel);
                      If Sel < 1 Then Sel := NumSelections;
                    End; { While }
                    SetSelection(Sel);
{                    SetSelection(CurSel);}
                  End;
        Key_Down: Begin
{                    Inc(CurSel);
                    If CurSel > NumSelections Then CurSel := 1;
                    SetSelection(CurSel);}
                    Sel := CurSel + 1;
                    While (Sel <> CurSel) And GetSelection(Sel)^.Protected Do
                    Begin
                      Inc(Sel);
                      If Sel > NumSelections Then Sel := 1;
                    End; { While }
                    SetSelection(Sel);
                  End;
        Key_Home: Begin
                    Sel := 1;
                    While (Sel < NumSelections) And GetSelection(Sel)^.Protected Do Inc(Sel);
                    SetSelection(Sel);
{                    SetSelection(1);}
                  End;
         Key_End: Begin
                    Sel := NumSelections;
                    While (Sel > 1) And GetSelection(Sel)^.Protected Do Dec(Sel);
                    SetSelection(Sel);
                  End;
        End; { Case }
      End;
      Else
      Begin
        FoundChar := False;
        SelWork   := SelHead;
        Sel       := 1;
        While Sel <= CurSel Do
        Begin
          Inc(Sel);
          SelWork := SelWork^.Next;
        End; { While }
        SelWork1 := SelWork;
        Repeat
          If (SelWork^.HotChar > 0) And
             (Key = UpCase(SelWork^.SelStr[SelWork^.HotChar])) And
             Not SelWork^.Protected Then FoundChar := True;
          SelWork := SelWork^.Next;
        Until (SelWork = SelWork1) Or FoundChar;
        If FoundChar Then
        Begin
          SetSelection(SelWork^.Last^.Index);
          Finished := True;
        End;
      End;
    End; { Case }
  End; { While }
  Remove;
End; { DMenu.Execute }


Procedure DMenu.Remove;
Begin
  OldSel := 0;
  Inherited Remove;
End; { DMenu.Remove }


Procedure ProtMenuInit;
Begin
  MaxX := Mem[$40:$4A];
  MaxY := Mem[$40:$84] + 1;
  If MaxY = 1 Then MaxY := 25;
End; { ProtMenuInit }

{------------------------------- }
{ Unit setup                     }
{------------------------------- }

Begin
  ProtMenuInit;
End.
